from django.core.exceptions import ImproperlyConfigured
from django.utils.unittest import skip

from ..models import Section

from ._utils import *
from arm_sections_support.models import *


class ModelTestCase(ArmSectionsTestCase):
    def setUp(self):
        super(ModelTestCase, self).setUp()

        self.pro_sports = Section.objects.get(slug='pro')
        self.weather = Section.objects.get(slug='weather')
        self.college = Section.objects.get(slug='college')

        self.complex_article = ComplexArticle.objects.create(
                title="Test Complex Article",
                slug='test_complex_article',
                primary_section=self.pro_sports,
            )
        self.complex_article.related_sections = [self.weather]

        self.multiple_many_to_many_article = MultipleManyToManyModel.objects.create(
                title="Test Multple Many To Many Article",
                slug='test_multiple_many_to_many_article',
                primary_section=self.pro_sports,
            )
        self.multiple_many_to_many_article.more_sections = [self.weather]

    def test_item_related_name_returns_many_to_many_field_name(self):
        with override_settings(ARMSTRONG_SECTION_ITEM_MODEL='armstrong.core.arm_sections.tests.arm_sections_support.models.ComplexCommon'):
            self.assertEqual(self.weather.item_related_name, 'related_sections')

    def test_item_related_name_returns_none_without_many_to_many(self):
        with override_settings(ARMSTRONG_SECTION_ITEM_MODEL='armstrong.core.arm_sections.tests.arm_sections_support.models.SimpleCommon'):
            self.assertEqual(self.weather.item_related_name, None)

    def test_item_related_name_returns_none_with_multiple_many_to_many(self):
        with override_settings(ARMSTRONG_SECTION_ITEM_MODEL='armstrong.core.arm_sections.tests.arm_sections_support.models.MultipleManyToManyModel'):
            self.assertEqual(self.weather.item_related_name, None)

    def test_choose_field_name_returns_specified_field(self):
        with override_settings(ARMSTRONG_SECTION_ITEM_MODEL='armstrong.core.arm_sections.tests.arm_sections_support.models.MultipleManyToManyModel'):
            self.assertEqual(self.weather._choose_field_name('related_sections'), 'related_sections')

    def test_choose_field_name_errors_with_multiple_many_to_many(self):
        with override_settings(ARMSTRONG_SECTION_ITEM_MODEL='armstrong.core.arm_sections.tests.arm_sections_support.models.MultipleManyToManyModel'):
            self.assertRaises(ImproperlyConfigured, self.weather._choose_field_name)

    def test_can_ensure_section(self):
        with override_settings(ARMSTRONG_SECTION_ITEM_MODEL='armstrong.core.arm_sections.tests.arm_sections_support.models.ComplexCommon'):
            self.assertNotIn(self.complex_article, self.college.items, msg='sanity check')
            self.college.add_item(self.complex_article)
            self.assertIn(self.complex_article, self.college.items)

    def test_can_ensure_section_with_field_name(self):
        with override_settings(ARMSTRONG_SECTION_ITEM_MODEL='armstrong.core.arm_sections.tests.arm_sections_support.models.MultipleManyToManyModel'):
            self.assertNotIn(self.multiple_many_to_many_article, self.college.items, msg='sanity check')
            self.college.add_item(self.multiple_many_to_many_article, field_name='more_sections')
            self.assertIn(self.multiple_many_to_many_article, self.college.items)

    def test_can_remove_section(self):
        with override_settings(ARMSTRONG_SECTION_ITEM_MODEL='armstrong.core.arm_sections.tests.arm_sections_support.models.ComplexCommon'):
            self.assertIn(self.complex_article, self.weather.items, msg='sanity check')
            self.weather.remove_item(self.complex_article)
            self.assertNotIn(self.complex_article, self.weather.items)

    def test_can_remove_section_with_field_name(self):
        with override_settings(ARMSTRONG_SECTION_ITEM_MODEL='armstrong.core.arm_sections.tests.arm_sections_support.models.MultipleManyToManyModel'):
            self.assertIn(self.multiple_many_to_many_article, self.weather.items, msg='sanity check')
            self.weather.remove_item(self.multiple_many_to_many_article, field_name='more_sections')
            self.assertNotIn(self.multiple_many_to_many_article, self.college.items)

    def test_can_toggle_section_with_true_test_func(self):
        with override_settings(ARMSTRONG_SECTION_ITEM_MODEL='armstrong.core.arm_sections.tests.arm_sections_support.models.ComplexCommon'):
            self.assertNotIn(self.complex_article, self.college.items, msg='sanity check')
            result = self.college.toggle_item(self.complex_article, lambda x: True)
            self.assertIn(self.complex_article, self.college.items)
            self.assertTrue(result)

    def test_can_toggle_section_with_true_test_func_and_field_name(self):
        with override_settings(ARMSTRONG_SECTION_ITEM_MODEL='armstrong.core.arm_sections.tests.arm_sections_support.models.MultipleManyToManyModel'):
            self.assertNotIn(self.multiple_many_to_many_article, self.college.items, msg='sanity check')
            result = self.college.toggle_item(self.multiple_many_to_many_article, lambda x: True, field_name='more_sections')
            self.assertIn(self.multiple_many_to_many_article, self.college.items)
            self.assertTrue(result)

    def test_can_toggle_section_with_false_test_func(self):
        with override_settings(ARMSTRONG_SECTION_ITEM_MODEL='armstrong.core.arm_sections.tests.arm_sections_support.models.ComplexCommon'):
            self.assertIn(self.complex_article, self.weather.items, msg='sanity check')
            result = self.weather.toggle_item(self.complex_article, lambda x: False)
            self.assertNotIn(self.complex_article, self.weather.items)
            self.assertFalse(result)

    def test_can_toggle_section_with_false_test_func_and_field_name(self):
        with override_settings(ARMSTRONG_SECTION_ITEM_MODEL='armstrong.core.arm_sections.tests.arm_sections_support.models.MultipleManyToManyModel'):
            self.assertIn(self.multiple_many_to_many_article, self.weather.items, msg='sanity check')
            result = self.weather.toggle_item(self.multiple_many_to_many_article, lambda x: False, field_name='more_sections')
            self.assertNotIn(self.multiple_many_to_many_article, self.weather.items)
            self.assertFalse(result)


class ManagerTestCase(ArmSectionsTestCase):
    def setUp(self):
        super(ManagerTestCase, self).setUp()

        self.pro_sports = Section.objects.get(slug='pro')
        self.weather = Section.objects.get(slug='weather')
        self.college = Section.objects.get(slug='college')

        self.complex_article = ComplexArticle.objects.create(
                title="Test Complex Article",
                slug='test_complex_article',
                primary_section=self.pro_sports,
            )
        self.complex_article.related_sections = [self.weather]

        self.multiple_many_to_many_article = MultipleManyToManyModel.objects.create(
                title="Test Multple Many To Many Article",
                slug='test_multiple_many_to_many_article',
                primary_section=self.pro_sports,
            )
        self.multiple_many_to_many_article.more_sections = [self.weather]

    def test_can_ensure_section_id(self):
        with override_settings(ARMSTRONG_SECTION_ITEM_MODEL='armstrong.core.arm_sections.tests.arm_sections_support.models.ComplexCommon'):
            self.assertNotIn(self.complex_article, self.college.items, msg='sanity check')
            Section.objects.add_item(self.complex_article, pk=self.college.pk)
            self.assertIn(self.complex_article, self.college.items)

    def test_can_ensure_section_id_with_field_name(self):
        with override_settings(ARMSTRONG_SECTION_ITEM_MODEL='armstrong.core.arm_sections.tests.arm_sections_support.models.MultipleManyToManyModel'):
            self.assertNotIn(self.multiple_many_to_many_article, self.college.items, msg='sanity check')
            Section.objects.add_item(self.multiple_many_to_many_article, field_name='more_sections', pk=self.college.pk)
            self.assertIn(self.multiple_many_to_many_article, self.college.items)

    def test_can_remove_section_id(self):
        with override_settings(ARMSTRONG_SECTION_ITEM_MODEL='armstrong.core.arm_sections.tests.arm_sections_support.models.ComplexCommon'):
            self.assertIn(self.complex_article, self.weather.items, msg='sanity check')
            Section.objects.remove_item(self.complex_article, pk=self.weather.pk)
            self.assertNotIn(self.complex_article, self.weather.items)

    def test_can_remove_section_id_with_field_name(self):
        with override_settings(ARMSTRONG_SECTION_ITEM_MODEL='armstrong.core.arm_sections.tests.arm_sections_support.models.MultipleManyToManyModel'):
            self.assertIn(self.multiple_many_to_many_article, self.weather.items, msg='sanity check')
            Section.objects.remove_item(self.multiple_many_to_many_article, field_name='more_sections', pk=self.weather.pk)
            self.assertNotIn(self.multiple_many_to_many_article, self.college.items)

    def test_can_toggle_section_id_with_true_test_func(self):
        with override_settings(ARMSTRONG_SECTION_ITEM_MODEL='armstrong.core.arm_sections.tests.arm_sections_support.models.ComplexCommon'):
            self.assertNotIn(self.complex_article, self.college.items, msg='sanity check')
            result = Section.objects.toggle_item(self.complex_article, lambda x: True, pk=self.college.pk)
            self.assertIn(self.complex_article, self.college.items)
            self.assertTrue(result)

    def test_can_toggle_section_id_with_true_test_func_and_field_name(self):
        with override_settings(ARMSTRONG_SECTION_ITEM_MODEL='armstrong.core.arm_sections.tests.arm_sections_support.models.MultipleManyToManyModel'):
            self.assertNotIn(self.multiple_many_to_many_article, self.college.items, msg='sanity check')
            result = Section.objects.toggle_item(self.multiple_many_to_many_article, lambda x: True, field_name='more_sections', pk=self.college.pk)
            self.assertIn(self.multiple_many_to_many_article, self.college.items)
            self.assertTrue(result)

    def test_can_toggle_section_id_with_false_test_func(self):
        with override_settings(ARMSTRONG_SECTION_ITEM_MODEL='armstrong.core.arm_sections.tests.arm_sections_support.models.ComplexCommon'):
            self.assertIn(self.complex_article, self.weather.items, msg='sanity check')
            result = Section.objects.toggle_item(self.complex_article, lambda x: False, pk=self.weather.pk)
            self.assertNotIn(self.complex_article, self.weather.items)
            self.assertFalse(result)

    def test_can_toggle_section_id_with_false_test_func_and_field_name(self):
        with override_settings(ARMSTRONG_SECTION_ITEM_MODEL='armstrong.core.arm_sections.tests.arm_sections_support.models.MultipleManyToManyModel'):
            self.assertIn(self.multiple_many_to_many_article, self.weather.items, msg='sanity check')
            result = Section.objects.toggle_item(self.multiple_many_to_many_article, lambda x: False, field_name='more_sections', pk=self.weather.pk)
            self.assertNotIn(self.multiple_many_to_many_article, self.weather.items)
            self.assertFalse(result)


from django.conf import settings
from django.utils.importlib import import_module


def get_module_and_model_names():
    s = (getattr(settings, "ARMSTRONG_SECTION_ITEM_MODEL", False) or
            "armstrong.apps.content.models.Content")
    return s.rsplit(".", 1)


def get_item_model_class():
    module_name, class_name = get_module_and_model_names()
    module = import_module(module_name)
    return getattr(module, class_name)

def filter_item_rels(rels):
    model_rels = []
    ItemModel = get_item_model_class()
    for related in rels:
        if issubclass(ItemModel, related.model):
            model_rels.append(related)
    return model_rels

def get_section_relations(Section):
    """Find every relationship between section and the item model."""
    all_rels = Section._meta.get_all_related_objects() + \
               Section._meta.get_all_related_many_to_many_objects()
    return filter_item_rels(all_rels)

def get_section_many_to_many_relations(Section):
    return filter_item_rels(Section._meta.get_all_related_many_to_many_objects())



from django.db.models import Q

from .utils import get_section_relations, get_item_model_class


class ItemFilter(object):
    manager_attr = 'objects'

    def get_manager(self, model):
        """Return the desired manager for the item model."""
        return getattr(model, self.manager_attr)

    def get_section_relations(self, section):
        return get_section_relations(section.__class__)

    def filter_objects_by_section(self, rels, section):
        """Build a queryset containing all objects in the section subtree."""
        subtree = section.get_descendants(include_self=True)
        kwargs_list = [{'%s__in' % rel.field.name: subtree} for rel in rels]
        q = Q(**kwargs_list[0])
        for kwargs in kwargs_list[1:]:
            q |= Q(**kwargs)
        return self.get_manager(get_item_model_class()).filter(q).distinct()

    def process_items(self, items):
        """
        Perform extra actions on the filtered items.

        Example: Further filtering the items in the section to meet a
        custom need.
        """
        if hasattr(items, 'select_subclasses'):
            items = items.select_subclasses()
        return items

    def __call__(self, section):
        relations = self.get_section_relations(section)
        items = self.filter_objects_by_section(relations, section)
        return self.process_items(items)


# Deprecated. This backend used to be a function that performed the same task.
find_related_models = ItemFilter()


class PublishedItemFilter(ItemFilter):
    manager_attr = 'published'



from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db import models
from mptt.models import MPTTModel
from mptt.fields import TreeForeignKey
from armstrong.utils.backends import GenericBackend

from .utils import get_section_many_to_many_relations

__BACKEND_MODULE = "armstrong.core.arm_sections.backends.%s"
SECTION_ITEM_BACKEND = (GenericBackend('ARMSTRONG_SECTION_ITEM_BACKEND',
        defaults=__BACKEND_MODULE % "ItemFilter")
    .get_backend())
SECTION_PUBLISHED_BACKEND = (GenericBackend(
        'ARMSTRONG_SECTION_PUBLISHED_BACKEND',
        defaults=__BACKEND_MODULE % "PublishedItemFilter")
    .get_backend())


class SectionManager(models.Manager):
    def get(self, **kwargs):
        defaults = {}
        defaults.update(kwargs)
        if 'full_slug' in defaults:
            if defaults['full_slug'] and defaults['full_slug'][-1] != "/":
                defaults['full_slug'] += "/"
        return super(SectionManager, self).get(**defaults)

    def add_item(self, item, field_name=None, **kwargs):
        section = self.get_query_set().get(**kwargs)
        section.add_item(item, field_name=field_name)

    def remove_item(self, item, field_name=None, **kwargs):
        section = self.get_query_set().get(**kwargs)
        section.remove_item(item, field_name=field_name)

    def toggle_item(self, item, test_func, field_name=None, **kwargs):
        section = self.get_query_set().get(**kwargs)
        return section.toggle_item(item, test_func, field_name=field_name)


class Section(MPTTModel):
    title = models.CharField(max_length=255)
    summary = models.TextField(default="", blank=True)
    slug = models.SlugField()
    full_slug = models.CharField(max_length=255, blank=True)

    parent = TreeForeignKey('self', null=True, blank=True)

    objects = SectionManager()

    @property
    def items(self):
        return SECTION_ITEM_BACKEND(self)

    @property
    def published(self):
        return SECTION_PUBLISHED_BACKEND(self)

    @property
    def item_related_name(self):
        """
        The ManyToMany field on the item class pointing to this class.

        If there is more than one field, this value will be None.
        """
        if not hasattr(self, '_item_related_name'):
            many_to_many_rels = \
                get_section_many_to_many_relations(self.__class__)
            if len(many_to_many_rels) != 1:
                self._item_related_name = None
            else:
                self._item_related_name = many_to_many_rels[0].field.name
        return self._item_related_name

    def save(self, *args, **kwargs):
        orig_full_slug = self.full_slug
        if self.parent:
            self.full_slug = "%s%s/" % (self.parent.full_slug, self.slug)
        else:
            self.full_slug = "%s/" % self.slug
        obj = super(Section, self).save(*args, **kwargs)
        if orig_full_slug != self.full_slug:
            for child in self.get_children():
                child.save()
        return obj

    def __unicode__(self):
        return "%s (%s)" % (self.title, self.full_slug)

    def _choose_field_name(self, specified_field=None):
        if specified_field is not None:
            return specified_field
        if self.item_related_name is None:
            raise ImproperlyConfigured(
                "A field_name must be specified if there isn't a single "
                "section ManyToMany relation.")
        return self.item_related_name

    def add_item(self, item, field_name=None):
        """
        Add the item to the specified section.

        Intended for use with items of settings.ARMSTRONG_SECTION_ITEM_MODEL.
        Behavior on other items is undefined.
        """
        field_name = self._choose_field_name(field_name)
        related_manager = getattr(item, field_name)
        related_manager.add(self)

    def remove_item(self, item, field_name=None):
        """
        Remove the item from the specified section.

        Intended for use with items of settings.ARMSTRONG_SECTION_ITEM_MODEL.
        Behavior on other items is undefined.
        """
        field_name = self._choose_field_name(field_name)
        related_manager = getattr(item, field_name)
        related_manager.remove(self)

    def toggle_item(self, item, test_func, field_name=None):
        """
        Toggles the section based on test_func.

        test_func takes an item and returns a boolean. If it returns True, the
        item will be added to the given section. It will be removed from the
        section otherwise.

        Intended for use with items of settings.ARMSTRONG_SECTION_ITEM_MODEL.
        Behavior on other items is undefined.
        """
        if test_func(item):
            self.add_item(item, field_name)
            return True
        else:
            self.remove_item(item, field_name)
            return False



from pkgutil import extend_path
__path__ = extend_path(__path__, __name__)

from .backends import *
from .models import *
from .template_tags import *
from .with_section import *
from .utils import *
from .views import *



from arm_sections_support.models import SimpleCommon

from ._utils import *

from .. import utils
from ..models import Section


def rel_field_names(rels):
    return [rel.field.name for rel in rels]


class get_configured_item_modelTestCase(ArmSectionsTestCase):
    def test_returns_configured_model(self):
        m = "%s.FooBar" % self.__class__.__module__
        with override_settings(ARMSTRONG_SECTION_ITEM_MODEL=m):
            module, model = utils.get_module_and_model_names()
            self.assertEqual(self.__class__.__module__, module)
            self.assertEqual("FooBar", model)

    def test_provides_default_value(self):
        with override_settings(ARMSTRONG_SECTION_ITEM_MODEL=False):
            module, model = utils.get_module_and_model_names()
            self.assertEqual("armstrong.apps.content.models", module)
            self.assertEqual("Content", model)


class get_item_model_classTestCase(ArmSectionsTestCase):
    def test_returns_specified_class(self):
        with override_settings(ARMSTRONG_SECTION_ITEM_MODEL='armstrong.core.arm_sections.tests.arm_sections_support.models.SimpleCommon'):
            self.assertEqual(SimpleCommon, utils.get_item_model_class())


class get_section_relationsTestCase(ArmSectionsTestCase):
    def test_returns_relation_for_foreign_key_only(self):
        with override_settings(ARMSTRONG_SECTION_ITEM_MODEL='armstrong.core.arm_sections.tests.arm_sections_support.models.SimpleCommon'):
            self.assertEqual(['primary_section'], rel_field_names(utils.get_section_relations(Section)))

    def test_returns_relations_for_foreign_key_and_many_to_many(self):
        with override_settings(ARMSTRONG_SECTION_ITEM_MODEL='armstrong.core.arm_sections.tests.arm_sections_support.models.ComplexCommon'):
            self.assertEqual(['primary_section', 'related_sections'], rel_field_names(utils.get_section_relations(Section)))

    def test_returns_relations_for_subclass_with_foreign_key_and_many_to_many(self):
        with override_settings(ARMSTRONG_SECTION_ITEM_MODEL='armstrong.core.arm_sections.tests.arm_sections_support.models.MultipleManyToManyModel'):
            self.assertEqual(['primary_section', 'related_sections', 'more_sections'], rel_field_names(utils.get_section_relations(Section)))


class get_section_many_to_many_relationsTestCase(ArmSectionsTestCase):
    def test_returns_no_relations_for_foreign_key_only(self):
        with override_settings(ARMSTRONG_SECTION_ITEM_MODEL='armstrong.core.arm_sections.tests.arm_sections_support.models.SimpleCommon'):
            self.assertEqual([], rel_field_names(utils.get_section_many_to_many_relations(Section)))

    def test_returns_relation_for_foreign_key_and_many_to_many(self):
        with override_settings(ARMSTRONG_SECTION_ITEM_MODEL='armstrong.core.arm_sections.tests.arm_sections_support.models.ComplexCommon'):
            self.assertEqual(['related_sections'], rel_field_names(utils.get_section_many_to_many_relations(Section)))

    def test_returns_relations_for_subclass_with_foreign_key_and_many_to_many(self):
        with override_settings(ARMSTRONG_SECTION_ITEM_MODEL='armstrong.core.arm_sections.tests.arm_sections_support.models.MultipleManyToManyModel'):
            self.assertEqual(['related_sections', 'more_sections'], rel_field_names(utils.get_section_many_to_many_relations(Section)))
